import { getTypes, createObject } from '@mo36924/graphql-schema';
import { escapeIdentifier, escapeLiteral } from '@mo36924/postgres-escape';
import { snakeCase } from 'change-case';

const data = (schema) => {
    const { Query, Mutation, ...types } = getTypes(schema);
    let sql = "";
    sql += disableForeignKeyCheck(types);
    sql += getInsertQueries(types);
    sql += enableForeignKeyCheck(types);
    return sql;
};
const disableForeignKeyCheck = (types) => Object.keys(types)
    .map((typeName) => `alter table ${escapeIdentifier(typeName)} disable trigger all;\n`)
    .join("");
const enableForeignKeyCheck = (types) => Object.keys(types)
    .map((typeName) => `alter table ${escapeIdentifier(typeName)} enable trigger all;\n`)
    .join("");
const uuid = (value, tableIndex) => "00000000-0000-4000-a000-" + tableIndex.toString().padStart(6, "0") + value.toString().padStart(6, "0");
const defaultScalarValues = createObject({
    ID: "",
    Int: 0,
    Float: 0,
    String: "",
    Boolean: true,
    Date: new Date(0),
});
const getInsertQueries = (types, baseRecordCount = 3) => {
    const tableIndexes = Object.fromEntries(Object.keys(types).map((typeName, index) => [typeName, index]));
    const recordCounts = createObject();
    const getRecordCount = (dep, deps = []) => {
        if (deps.includes(dep)) {
            return 0;
        }
        if (recordCounts[dep]) {
            return recordCounts[dep];
        }
        const fields = types[dep].fields;
        const recordCount = Math.max(baseRecordCount, ...Object.values(fields).map(({ directives: { ref, unique } }) => ref ? getRecordCount(ref.name, [dep, ...deps]) * (unique ? 1 : baseRecordCount) : 0));
        recordCounts[dep] = recordCount;
        return recordCount;
    };
    const getInsertQuery = (typeName) => {
        const recordCount = getRecordCount(typeName);
        const fieldEntries = Object.entries(types[typeName].fields).filter(([, field]) => field.scalar);
        const fieldNames = fieldEntries.map(([fieldName]) => escapeIdentifier(fieldName));
        const valuesList = [];
        for (let i = 0; i < recordCount; i++) {
            const values = [];
            for (const [fieldName, field] of fieldEntries) {
                const { type: fieldType, directives: { ref }, } = field;
                const defaultValue = defaultScalarValues[fieldType];
                let value;
                if (fieldName === "id") {
                    value = uuid(i + 1, tableIndexes[typeName]);
                }
                else if (fieldName === "version") {
                    value = 1;
                }
                else if (ref) {
                    value = uuid((i % getRecordCount(ref.name)) + 1, tableIndexes[typeName]);
                }
                else if (typeof defaultValue === "string") {
                    value = `${fieldName}-${i + 1}`;
                }
                else {
                    value = defaultValue;
                }
                values.push(escapeLiteral(value));
            }
            valuesList.push(`(${values.join()})`);
        }
        return `insert into ${escapeIdentifier(typeName)} (${fieldNames.join()}) values \n${valuesList.join(",\n")};\n`;
    };
    return Object.keys(types).map(getInsertQuery).join("");
};

const dbTypes = createObject({
    UUID: "uuid",
    ID: "uuid",
    Int: "integer",
    Float: "double precision",
    String: "text",
    Boolean: "boolean",
    Date: "timestamp(3)",
});
const schema = (schema) => {
    const { Query, Mutation, ...types } = getTypes(schema);
    const extension = [`create extension if not exists pgcrypto;\n`];
    const create = [];
    const unique = [];
    const index = [];
    const key = [];
    const trigger = [];
    for (const [typeName, type] of Object.entries(types)) {
        const { fields } = type;
        const columns = [];
        for (const [fieldName, field] of Object.entries(fields)) {
            const { type: fieldTypeName, list, null: _null, scalar, directives: { ref: refDirective, unique: uniqueDirective }, } = field;
            if (scalar) {
                const dbType = dbTypes[fieldTypeName];
                switch (fieldName) {
                    case "id":
                        columns.push(`${escapeIdentifier(fieldName)} ${dbType} not null default gen_random_uuid() primary key`);
                        break;
                    case "version":
                        columns.push(`${escapeIdentifier(fieldName)} ${dbType} not null default 1`);
                        break;
                    case "createdAt":
                        columns.push(`${escapeIdentifier(fieldName)} ${dbType} not null default current_timestamp(3)`);
                        break;
                    case "updatedAt":
                        columns.push(`${escapeIdentifier(fieldName)} ${dbType} not null default current_timestamp(3)`);
                        trigger.push(`create trigger update_at_${snakeCase(typeName)} before update on ${escapeIdentifier(typeName)} for each row execute procedure update_at();\n`);
                        break;
                    default:
                        columns.push(`${escapeIdentifier(fieldName)} ${dbType}${_null ? "" : " not null"}`);
                        break;
                }
            }
            if (uniqueDirective) {
                unique.push(`alter table ${escapeIdentifier(typeName)} add unique (${escapeIdentifier(fieldName)});\n`);
            }
            else if (refDirective) {
                index.push(`create index on ${escapeIdentifier(typeName)} (${escapeIdentifier(fieldName)});\n`);
            }
            if (refDirective) {
                key.push(`alter table ${escapeIdentifier(typeName)} add foreign key (${escapeIdentifier(fieldName)}) references ${escapeIdentifier(refDirective.name)} (${escapeIdentifier("id")});\n`);
            }
        }
        create.push(`create table ${escapeIdentifier(typeName)} (\n${columns.map((column) => `  ${column}`).join(",\n")}\n);\n`);
    }
    if (trigger.length) {
        trigger.unshift(`create or replace function update_at() returns trigger as $$ begin new."updatedAt" := current_timestamp(3); return new; end; $$ language plpgsql;\n`);
    }
    return [extension, create, unique, index, key, trigger].flat().join("");
};

export { data, schema };
//# sourceMappingURL=index.mjs.map
